@@ -64,13 +64,12 @@ static char other_device [512];
 static int reboot = 0;
 static int floppyswap = 0;
 int tab_ambiguous = 0;
-unsigned int linux_end = 0;
 int other_part = -1;
 int solaris = 0;
 int other = 0;
 char *password = 0;
 
-void parse_name (char *, int, char **, int *, char **);
+static void parse_name (char *, int, char **, int *, char **);
 
 static char *next_tok (char *t) {
     while (*t++); /* That was easy */
@@ -173,7 +172,8 @@ static void maintabfunc (void)
     return;
 }
 
-void parse_name (char *imagename, int defpart, char **device, int *part, char **kname)
+static void parse_name (char *imagename, int defpart, char **device,
+			int *part, char **kname)
 {
     static char parsebuf[1024];
 
@@ -231,7 +231,7 @@ void parse_name (char *imagename, int defpart, char **device, int *part, char **
 	*kname = 0;
 }
 
-void check_initrd (char *label, int defpart, char *defdevice)
+static void check_initrd (char *label, int defpart, char *defdevice)
 {
     char *p;
     
@@ -246,7 +246,7 @@ void check_initrd (char *label, int defpart, char *defdevice)
     }
 }
 
-int dig_into_params (char *params)
+static int dig_into_params (char *params)
 {
     char *p, *q;
     char *last_root = 0;
@@ -300,7 +300,7 @@ int dig_into_params (char *params)
     return root_found;
 }
 
-void check_password(char *str)
+static void check_password(char *str)
 {
     int i;
     
@@ -318,7 +318,7 @@ void check_password(char *str)
     prom_halt();
 }
 
-char *ls_options(char *parms)
+static char *ls_options(char *parms)
 {
     char *p = parms;
     ls_opt = 0;
@@ -338,7 +338,8 @@ char *ls_options(char *parms)
     return p;
 }
 
-int get_params (char **device, int *part, char **kname, char **proll, char **params)
+static int get_params (char **device, int *part, char **kname, char **proll,
+		       char **params)
 {
     int defpart = -1;
     char *defdevice = 0;
@@ -831,21 +832,24 @@ int get_params (char **device, int *part, char **kname, char **proll, char **par
     return 0;
 }
 
-void initrd_lenfunc (int len, char **filebuffer, char **filelimit)
+static void initrd_lenfunc (int len, char **filebuffer, char **filelimit)
 {
+    extern unsigned long long initrd_phys;
+
     initrd_start = memory_find ((len + 16383) & ~16383);
     if (!initrd_start) {
         fatal ("You do not have enough continuous available memory for such initial ramdisk.");
         prom_halt ();
     }
     initrd_size = len;
-    printf ("Loading initial ramdisk....\n");
     *filebuffer = initrd_start;
     *filelimit = initrd_start + ((len + 16383) & ~16383);
+    printf("Loading initial ramdisk (%d bytes at 0x%Lx phys, 0x%x virt)...\n", len,
+	   initrd_phys, initrd_start);
 }
 
-int parse_executable (char *base, int image_len, unsigned int *poff, int *plen,
-    unsigned *pstart, char *image_name)
+static int parse_executable (char *base, int image_len, unsigned int *poff,
+			     int *plen, unsigned *pstart, char *image_name)
 {
     int isfile = 0;
     union {
@@ -1025,8 +1029,6 @@ int bootmain (void)
     strcpy (given_bootargs, get_bootargs (1));
     strcpy (my_bootargs, get_bootargs (0));
 
-    printf (" ");
-
 #ifndef TFTP
     if (diskinit () == -1)
 	prom_halt ();
@@ -1058,7 +1060,9 @@ int bootmain (void)
     if (*silo_conf && silo_conf_partition >= 1 && silo_conf_partition <= 8) {
 	int len;
 	solaris = 0;
-	fileok = load_file (0, silo_conf_partition, silo_conf, (unsigned char *) 0x4000, (unsigned char *) &_start, &len, LOADFILE_GZIP, 0);
+	fileok = load_file (0, silo_conf_partition, silo_conf,
+			    (unsigned char *) 0x4000, (unsigned char *) &_start,
+			    &len, LOADFILE_GZIP | LOADFILE_NO_ROTATE, 0);
 	if (!fileok || (unsigned) len >= 65535)
 	    printf ("\nCouldn't load %s\n", silo_conf);
 	else {
@@ -1106,7 +1110,7 @@ int bootmain (void)
 	"Type [prompath;]part/path_to_image [parameters] on the prompt\n"
 	  "E.g. /iommu/sbus/espdma/esp/sd@3,0;4/vmlinux root=/dev/sda4\n"
 		"or 2/vmlinux.live (to load vmlinux.live from 2nd partition of boot disk)\n");
-
+try_again:
     isfile = 0;			/* RC = 0 invalid file or not an executable */
     while (!isfile) {
 	switch (get_params (&device, &part, &kname, &proll, &params)) {
@@ -1159,11 +1163,36 @@ int bootmain (void)
 	    ret_offset = 0x4000;
 
 	} else {
-	    image_base = (unsigned char *) 0x4000;
-	    if (!load_file (device, part, kname, image_base,
-			(unsigned char *) &_start, &image_len,
-			load_cmd == CMD_LS ? LOADFILE_LS : LOADFILE_GZIP, 0)) {
+	    char *image_end = (char *)&_start;
+
+	    /* See if we can use some extra memory for the kernel */
+	    if (!load_cmd) {
+		unsigned int size;
+		char *mem;
+
+		size = 0x800000;
+		mem = image_memory_find(size);
+
+		if (!mem) {
+		    size = 0x400000;
+		    mem = image_memory_find(size);
+		}
+
+		if (mem) {
+		    image_base = mem;
+		    image_end = image_base + size - 0x4000;
+		    printf("Allocated %d Megs of memory at 0x%x for kernel\n",
+			   size >> 20, image_base - 0x4000);
+		}
+	    }
+
+	    if (!load_file (device, part, kname, image_base, image_end,
+			&image_len, load_cmd == CMD_LS ? LOADFILE_LS : LOADFILE_GZIP, 0)) {
 		printf ("\nImage not found.... try again\n");
+
+		if (!load_cmd)
+		    image_memory_release();
+
         	continue;
             }
 
@@ -1178,8 +1207,8 @@ int bootmain (void)
         	continue;
 	    }
 
-            isfile = parse_executable (image_base, image_len, &off, &len,
-		&ret_offset, kname);
+	    isfile = parse_executable (image_base, image_len, &off, &len,
+				       &ret_offset, kname);
 	}
     }
 
@@ -1188,11 +1217,50 @@ int bootmain (void)
     	params = params_device;
     	params_device = sol_params;
     } else if (!other) {
+	unsigned short hdrs_ver = 0;
+
     	params_device = 0;
-        memcpy (image_base, image_base + off, len);
+
+	memcpy (image_base, image_base + off, len);
+
         p = find_linux_HdrS (image_base, image_len);
+
+	if (p)
+	    hdrs_ver = *(unsigned short *)(p + 8);
+
+	if (p && hdrs_ver < 0x300 && image_base != (char *)0x4000) {
+	    /* Kernel doesn't support being loaded to other than
+	     * phys_base, so let's try to copy it down there. */
+	    if ((unsigned int)&_start - 0x4000 < len) {
+		/* Fuck, can't do that */
+		printf("Your kernel cannot fit into the memory destination. This\n"
+		       "can be resolved by recompiling the kernel with more devices\n"
+		       "built as modules, or upgrading your kernel to one that\n"
+		       "supports being loaded to higher memory areas (currently\n"
+		       "2.6.3+ or 2.4.26+).\n");
+		goto try_again;
+	    }
+
+	    printf("Kernel doesn't support loading to high memory, relocating...");
+
+	    /* Ok, it fits, so copy it down there */
+	    memcpy ((char *)0x4000, image_base, len);
+	    image_base = (char *)0x4000;
+
+	    /* Readjust some things */
+	    ret_offset = 0x4000;
+	    p = find_linux_HdrS (image_base, image_len);
+
+	    printf("done.\n");
+	}
+
         if (p) {
-            if (fill_reboot_cmd && *(unsigned short *)(p + 8) >= 0x201) { /* ie. uses reboot_command */
+	    unsigned int linux_version = *(unsigned int *)(p + 4);
+
+	    printf("Loaded kernel version %d.%d.%d\n", (linux_version >> 16) & 0xff,
+		   (linux_version >> 8) & 0xff, linux_version & 0xff);
+
+            if (fill_reboot_cmd && hdrs_ver >= 0x201) { /* ie. uses reboot_command */
                 char *q = (char *)(*(unsigned int *)(p + 24)), *r;
                 extern char bootdevice[];
 
@@ -1228,12 +1296,15 @@ int bootmain (void)
                     params = s2;
                 }
             }
-            if (*(unsigned short *)(p + 8) >= 0x202)
-            	kernel_params = (char *)(*(unsigned int *)(p + 36)&0x3fffff);
-            if (*(unsigned short *)(p + 8) >= 0x203) {
-		linux_end = *(unsigned int *)(p + 40)&0x3fffff;
-		linux_end = (linux_end + 32 + 0x1fff) & ~0x1fff;
+
+            if (hdrs_ver >= 0x202) {
+		if (architecture == sun4u)
+		    kernel_params = (char *)((*(unsigned int *)(p + 36) - 0x400000) + 
+				(image_base - 0x4000));
+		else
+		    kernel_params = (char *)(*(unsigned int *)(p + 36) & 0x3fffff);
 	    }
+
             /* Some UltraAX machines have /dev/fd1 floppies only. */
             if (floppyswap) {
             	char *s1;
@@ -1241,6 +1312,7 @@ int bootmain (void)
             	for (s1 = params; (s1 = strstr(s1, "root=/dev/fd0")) != NULL; s1 += 13)
             		s1[12] = '1';
             }
+
             if (initrd_string) {
             	char *q, *r, *initrd_device, *initrd_kname, *initrd_limit, *initrd_cur, c;
 		char *string;
@@ -1272,11 +1344,11 @@ int bootmain (void)
 			if (!initrd_kname) break;
 			if (!initrd_device) initrd_device = initrd_defdevice;
 			if (!load_file (initrd_device, initrd_partno, initrd_kname, initrd_cur, initrd_limit, &len, 0, 0)) break;
+			initrd_cur += len;
 			if (!c) {
 			    statusok = 1;
 			    break;
 			}
-			initrd_cur += len;
 			r = q + 1;
 			q = strchr (r, '|');
 			if (!q) q = strchr (r, 0);
@@ -1288,13 +1360,15 @@ int bootmain (void)
 			}
     		    }
     		    free (string);
+		    printf("Loaded initial ramdisk (%d bytes at 0x%x)...\n", (unsigned int)initrd_cur -
+			   (unsigned int)initrd_start, initrd_start);
     		    if (statusok) {
-	        	extern unsigned long sun4u_initrd_pa;
+	        	extern unsigned long long sun4u_initrd_phys;
 	        	extern unsigned long sun4m_initrd_pa;
-	        	if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
-			    *(unsigned int *)(p + 16) = linux_end;
-			else if (architecture == sun4u)
-	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
+
+			if (architecture == sun4u)
+	            	    *(unsigned int *)(p + 16) = (unsigned int)sun4u_initrd_phys +
+				    (hdrs_ver < 0x301 ? 0x400000 : 0);
 	            	else if (sun4m_initrd_pa)
 	            	    *(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
 	            	else
@@ -1306,20 +1380,23 @@ int bootmain (void)
 		    parse_name (initrd_string, initrd_defpart, &initrd_device, &initrd_partno, &initrd_kname);
 		    if (initrd_kname) {
 		        if (!initrd_device) initrd_device = initrd_defdevice;
-	        	if (load_file (initrd_device, initrd_partno, initrd_kname, (unsigned char *) 0x300000, (unsigned char *) 0x380000, 0, 0, initrd_lenfunc)) {
-	        	    extern unsigned long sun4u_initrd_pa;
+	        	if (load_file (initrd_device, initrd_partno, initrd_kname,
+				       (char *) 0x300000, (unsigned char *) LARGE_RELOC,
+				       0, 0, initrd_lenfunc)) {
+	        	    extern unsigned long long sun4u_initrd_phys;
 	        	    extern unsigned long sun4m_initrd_pa;
-			    if ((linux_end & 0x3fffff) == (unsigned int)initrd_start)
-				*(unsigned int *)(p + 16) = linux_end;
-			    else if (architecture == sun4u)
-	            	        *(unsigned int *)(p + 16) = ((unsigned int)sun4u_initrd_pa + 0x400000);
+
+			    if (architecture == sun4u)
+	            	        *(unsigned int *)(p + 16) = (unsigned int)sun4u_initrd_phys + 0x400000;
 			    else if (sun4m_initrd_pa)
 				*(unsigned int *)(p + 16) = ((unsigned int)sun4m_initrd_pa);
 	            	    else
 	            	        *(unsigned int *)(p + 16) = ((unsigned int)initrd_start | 0xf0000000);
+
 	            	    *(unsigned int *)(p + 20) = initrd_size;
 	        	}
-		    }
+		    } else
+			printf ("Error: initial ramdisk loading failed. No initrd will be used.\n");
 		}
 	    }
         }
@@ -1332,6 +1409,7 @@ int bootmain (void)
         strcpy (other_device, p);
         params_device = other_device;
     }
+
     close ();
     if (timer_status >= 1)
         close_timer ();
@@ -1343,6 +1421,7 @@ int bootmain (void)
 	}
     } else {
 	set_bootargs (params, params_device);
+
 	if (kernel_params) {
     	    extern char barg_out[];
     	    int len = *(unsigned int *)kernel_params;
@@ -1351,22 +1430,27 @@ int bootmain (void)
     	    kernel_params [8 + len - 1] = 0;
     	    *(unsigned int *)(kernel_params + 4) = 1;
 	}
+
 	if (show_arguments) {
     	    show_bootargs ();
     	    pause_after = 1;
 	}
     }
+
     if (pause_after) {
         printf ("%s", pause_message);
         prom_getchar ();
         printf ("\n");
     }
+
     memory_release();
+
     if (other && reboot) {
         strcpy (sol_params, params_device);
         strcat (sol_params, " ");
         strcat (sol_params, params);
     	prom_reboot(sol_params);
     }
+
     return ret_offset;
 }
