@@ -50,7 +50,8 @@
 #endif
 #ifdef __linux__
 #  include <linux/fs.h>
-#  include <linux/ext2_fs.h>
+#  include <ext2fs/ext2_fs.h>
+#  include <ext2fs/ext2fs.h>
 #  include <scsi/scsi.h>
 #  include <sys/vfs.h>
 #  ifdef __GLIBC__
@@ -89,6 +90,7 @@ static int ufs_blocks (struct hwdevice *, ino_t);
 #include "../first/ultra.h"
 #include "../first/fd.h"
 #include "../first/ieee32.h"
+
 #define DIGIT_OFFSET (DIGIT_OFFSET_TMP + 0x223)
 #define LETTER_OFFSET (LETTER_OFFSET_TMP + 0x223)
 #define NUMBER_OFFSET (NUMBER_OFFSET_TMP + 0x223)
@@ -99,6 +101,8 @@ static int ufs_blocks (struct hwdevice *, ino_t);
 
 /* Checks our config file for errors */
 int confcheck(char*);
+char *cfg_get_strg (char *, char *);
+int cfg_get_flag (char *, char *);
 static int allow_confchk_fail = 0;
 
 /* This is just so that we don't have to fight with incompatible ufs_fs.h headers */
@@ -242,7 +246,7 @@ void read_sb (struct hwdevice *hwdev)
 	fatal ("Cannot open superblock on %s", hwdev->dev);
     hwdev->bs = check_fs (fd);
     if (hwdev->bs == (unsigned short)-1)
-	fatal ("File systems other than ext2, ufs and romfs not yet supported", hwdev->dev);
+	fatal ("File systems other than ext2, ext3, ufs and romfs not yet supported", hwdev->dev);
     close (fd);
     hwdev->nsect = hwdev->bs / 512;
     if (hwdev->part == -1)
@@ -672,10 +676,9 @@ void flash_check_sum (char *data, unsigned int dsize)
 {
     unsigned int sum;
 
-    sum = 0;
-    while (dsize-- != 0) {
-        sum += *((unsigned char *)data)++;
-    }
+    for (sum = 0; dsize-- != 0; data++)
+        sum += *((unsigned char *)data);
+
     data[0] = sum >> 24;
     data[1] = sum >> 16;
     data[2] = sum >> 8;
@@ -1002,17 +1005,19 @@ struct hwdevice *get_device(int majno, int minno)
 				md_disk_info.number = i;
 				if (ioctl (md_fd, GET_DISK_INFO, &md_disk_info) < 0)
 					fatal ("Could not get RAID disk info for disk %d\n", i);
-				d = get_device (md_disk_info.majorno, md_disk_info.minorno);
-				if (md_disk_info.state == MD_DISK_FAULTY) {
-					printf ("disk %s marked as faulty, skipping\n", d->dev);
-					continue;
+				if(md_disk_info.majorno != 0 && md_disk_info.minorno != 0) {
+					d = get_device (md_disk_info.majorno, md_disk_info.minorno);
+					if (md_disk_info.state == MD_DISK_FAULTY) {
+						printf ("disk %s marked as faulty, skipping\n", d->dev);
+						continue;
+					}
+					if (hwdev)
+						last->next = d;
+					else
+						hwdev = d;
+					while (d->next != NULL) d = d->next;
+					last = d;
 				}
-				if (hwdev)
-					last->next = d;
-				else
-					hwdev = d;
-				while (d->next != NULL) d = d->next;
-				last = d;
 			}
 			if (!hwdev)
 				fatal ("No non-faulty disks found in RAID1");
@@ -1058,7 +1063,7 @@ struct hwdevice *get_device(int majno, int minno)
 
 int main(int argc,char **argv)
 {
-    char *name = NULL, *config_file, *install = NULL, *secondary, *backup, *p;
+    char *name = NULL, *config_file, *install = NULL, *secondary = NULL, *backup, *p;
     struct utsname uts;
     int version = 0;
     struct stat st1, st2, st3;
@@ -1078,7 +1083,6 @@ int main(int argc,char **argv)
     	ultra = 1;
     config_file = DFL_CONFIG;
     backup = DFL_BACKUP;
-    secondary = DFL_SECONDARY;
     new_root = NULL;
     name = *argv++;
     argc--;
@@ -1222,6 +1226,17 @@ int main(int argc,char **argv)
     if (!confcheck(config_file) && !allow_confchk_fail)
 	return 1;
 
+    /* Check for some values in the config that override defaults */
+    if (cfg_get_flag(NULL, "partition-boot"))
+	masterboot = 0;
+
+    if (!secondary) {
+	if (cfg_get_strg(NULL, "secondary"))
+	    secondary = cfg_get_strg(NULL, "secondary");
+	else
+	    secondary = DFL_SECONDARY;
+    }
+
     secondary = strdup (secondary);
     if (stat (secondary, &st1) < 0)
         fatal ("Cannot open second stage loader %s", secondary);
